from .api_client import ApiClient, DeviceInfo
from .auth_client import DeebotAuthClient
from .credentials import Credentials
from .entity import DeebotEntity
from .portal_client import PortalClient
from .subscription_client import SubscriptionClient

__all__ = [
    "ApiClient",
    "DeviceInfo",
    "DeebotAuthClient",
    "Credentials",
    "DeebotEntity",
    "PortalClient",
    "SubscriptionClient",
]
__version__ = "1.0.2"



import datetime
import logging
from typing import NamedTuple

from deebot_t8.auth_client import Authenticator
from deebot_t8.exceptions import ApiErrorException
from deebot_t8.portal_client import PortalClient
from deebot_t8.urls import APP_DO_PATH, DEVMANAGER_DO_PATH

LOGGER = logging.getLogger(__name__)


class DeviceInfo(NamedTuple):
    id: str
    id_short: str
    name: str
    product_category: str
    model_code: str
    model_name: str
    status: int

    dev_class: str
    resource: str

    icon_url: str


class ApiClient:
    def __init__(
        self,
        portal_client: PortalClient,
        authenticator: Authenticator,
    ):
        self._portal_client = portal_client
        self._authenticator = authenticator

    def get_devices_list(self):
        credentials = self._authenticator.authenticate()
        resp = self._portal_client.do_post(
            APP_DO_PATH,
            {
                "userid": credentials.user_id,
                "todo": "GetGlobalDeviceList",
            },
            credentials=credentials,
        )

        rv = []
        for device in resp["devices"]:
            rv.append(
                DeviceInfo(
                    id=device["did"],
                    id_short=device["name"],
                    name=device["nick"],
                    product_category=device["product_category"],
                    model_code=device["model"],
                    model_name=device["deviceName"],
                    status=device["status"],
                    dev_class=device["class"],
                    resource=device["resource"],
                    icon_url=device["icon"],
                )
            )

        return rv

    def exc_command(self, recipient: DeviceInfo, command: str, data=None):
        credentials = self._authenticator.authenticate()
        payload = {
            "header": {
                "pri": "2",
                "ts": datetime.datetime.now().timestamp(),
                "tmz": 480,
                "ver": "0.0.22",
            }
        }
        if data is not None:
            payload["body"] = {
                "data": data,
            }

        resp = self._portal_client.do_post(
            DEVMANAGER_DO_PATH,
            {
                "cmdName": command,
                "payload": payload,
                "payloadType": "j",
                "td": "q",
                "toId": recipient.id,
                "toRes": recipient.resource,
                "toType": recipient.dev_class,
            },
            query={
                "mid": recipient.dev_class,
                "did": recipient.id,
                "td": "q",
                "u": credentials.user_id,
                "cv": "1.67.3",
                "t": "a",
                "av": "1.3.1",
            },
            credentials=credentials,
        )
        if resp["ret"] != "ok":
            raise ApiErrorException(resp)

        if resp["resp"]["body"]["code"] != 0:
            raise ApiErrorException(resp)

        return resp["resp"]



import json
import logging
import re
import ssl
import threading
from typing import Any, Tuple, Callable, Dict, Set

from paho.mqtt.client import Client as MQTTClient, MQTT_ERR_SUCCESS

from .api_client import DeviceInfo
from .auth_client import Authenticator
from .urls import REALM

TOPIC_RE = re.compile(
    "iot/atr/(?P<command>[^/]+)/(?P<device_id>[^/]+)/(?P<device_cls>[^/]+)/"
    "(?P<device_resource>[^/]+)/j"
)

LOGGER = logging.getLogger(__name__)

HandlerT = Callable[[str, Dict[str, Any]], None]


class SubscriptionClient:
    def __init__(self, authenticator: Authenticator, continent: str, device_id: str):
        self._authenticator = authenticator
        self._continent = continent
        self._device_id = device_id

        self._lock = threading.Lock()
        # (device id, handler(command, data))
        self._subscribers: Set[Tuple[DeviceInfo, HandlerT]] = set()
        self._client = None

    def _connect(self):
        LOGGER.debug("MQTT Connecting...")

        credentials = self._authenticator.authenticate()
        client_id = (
            f"{credentials.user_id}@{REALM.split('.')[0]}/{self._device_id[0:8]}"
        )
        username = f"{credentials.user_id}@{REALM}"
        password = credentials.access_token
        url = self._get_server_url()

        LOGGER.debug("url: %s", url)
        LOGGER.debug("client_id: %s", client_id)
        LOGGER.debug("username: %s password: %s", username, password)

        self._client = MQTTClient(client_id=client_id)

        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        self._client.username_pw_set(username, password)

        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE
        self._client.tls_set_context(ssl_ctx)
        self._client.tls_insecure_set(True)

        self._client.connect(url, port=8883)
        self._client.loop_start()

    def _on_connect(self, client, userdata, flags, rc):
        if rc != MQTT_ERR_SUCCESS:
            LOGGER.error("MQTT Connection failure. rc: %d", rc)
            # TODO(NW): Handle specific auth failure, e.g. auth token expired
            #  MQTT_ERR_AUTH perhaps?
            return

        LOGGER.debug("Connected with result code " + str(rc))
        for (device, _) in self._subscribers:
            self._subscribe_topic(device)

        self._client.subscribe("#", qos=0)

    def _on_message(self, client, userdata, msg):
        LOGGER.debug("mqtt: message received: %s", msg.topic)

        if m := TOPIC_RE.match(msg.topic):
            command = m.group("command")
            device_id = m.group("device_id")

            for (device, handler) in self._subscribers:
                if device.id == device_id:
                    data = json.loads(msg.payload)
                    handler(command, data["body"])

    def _get_server_url(self):
        """
        US 	mq-na.ecouser.net
        "World-wide" mq-ww.ecouser.net
        """
        return f"mq-{self._continent}.ecouser.net"

    def subscribe(self, device: DeviceInfo, handler):
        with self._lock:
            self._subscribers.add((device, handler))

            if self._client is None:
                self._connect()
                return

            if self._client.is_connected:
                self._subscribe_topic(device)

    def unsubscribe(self, device: DeviceInfo, handler):
        to_remove = []
        with self._lock:
            for elem in self._subscribers:
                (s_device, s_handler) = elem
                if s_device.id == device.id and s_handler == handler:
                    to_remove.append(elem)

            for entry in to_remove:
                self._subscribers.remove(entry)

            if len(self._subscribers) == 0:
                LOGGER.info("No remaining subscribers, disconnecting from MQTT")
                if self._client is not None:
                    self._client.disconnect()
                    self._client = None

    def _subscribe_topic(self, device: DeviceInfo):
        topic = (
            "iot/atr/+/"
            + device.id
            + "/"
            + device.dev_class
            + "/"
            + device.resource
            + "/+"
        )
        LOGGER.debug("Subscribing to topic: {}".format(topic))
        if self._client is None:
            raise AssertionError("client is null")

        self._client.subscribe(topic, qos=0)



from __future__ import annotations
import json
import logging
import threading
import time
from dataclasses import dataclass
from enum import Enum
from typing import List, Callable, Set

from deebot_t8.api_client import ApiClient, DeviceInfo
from deebot_t8.exceptions import ApiErrorException
from deebot_t8.subscription_client import SubscriptionClient

LOGGER = logging.getLogger(__name__)


@dataclass
class ComponentLifeSpan:
    component: str
    total: int
    left: int


@dataclass
class TotalStats:
    area: int
    time: int
    count: int


@dataclass
class CleanStats:
    area: int
    time: int
    avoid_count: int
    start_time: int


@dataclass
class VacuumState:
    class Speed(Enum):
        QUIET = 1
        STANDARD = 2
        MAX = 3
        MAX_PLUS = 4

    class WaterFlow(Enum):
        LOW = 1
        MEDIUM = 2
        HIGH = 3
        ULTRA_HIGH = 4

    class RobotState(Enum):
        IDLE = 1
        CLEANING = 2
        PAUSED = 3
        RETURNING = 4

    class CleanType(Enum):
        AUTO = 1
        SPOT_AREA = 2
        CUSTOM_AREA = 3

    is_online: bool | None = None
    firmware_version: str | None = None
    hardware_version: str | None = None

    state: RobotState | None = None
    clean_type: CleanType | None = None
    clean_stats: CleanStats | None = None

    battery_level: int | None = None
    is_charging: bool | None = None

    mop_attached: bool | None = None
    water_level: WaterFlow | None = None

    vacuum_speed: Speed | None = None
    clean_count: int | None = None

    cleaning_preference_enabled: bool | None = None
    true_detect_enabled: bool | None = None
    auto_boost_suction_enabled: bool | None = None
    auto_empty_enabled: bool | None = None

    lifespan: List[ComponentLifeSpan] | None = None
    total_stats: TotalStats | None = None

    _on_change: Callable[[VacuumState, str], None] | None = None

    def __setattr__(self, key, value):
        super(VacuumState, self).__setattr__(key, value)
        if getattr(self, key) != value:
            LOGGER.debug("set state: {} -> {}".format(key, value))

        if self._on_change is not None:
            self._on_change(self, key)


class DeebotEntity:
    def __init__(
        self, api_client: ApiClient, subs_client: SubscriptionClient, device: DeviceInfo
    ):
        self._api_client = api_client
        self._subs_client = subs_client
        self._device = device

        self._lock = threading.Lock()
        self._subscribers: Set[Callable[[VacuumState, str], None]] = set()

        self.state: VacuumState = VacuumState(is_online=device.status == 1)
        self.state._on_change = self._handle_state_change

        self._should_poll = False
        self._err_count = 0

    def force_refresh(self):
        requests = [
            (
                "getInfo",
                [
                    "getCleanInfo_V2",
                    "getWaterInfo",
                    "getChargeState",
                    "getBattery",
                    "getStats",
                    "getError",
                ],
            ),
            # Make an additional getInfo call to avoid response payload size
            # restriction. It appears larger response payloads are returned via
            # MQTT across multiple messages. This saves additional
            # implementation complexity to support that.
            (
                "getInfo",
                [
                    "getTotalStats",
                    "getSpeed",
                    "getCleanCount",
                    "getTrueDetect",
                    "getCleanPreference",
                    "getAutoEmpty",
                    "getCarpertPressure",  # <- Yes there is actually a typo in the api 🤷‍
                ],
            ),
            ("getLifeSpan", ["sideBrush", "brush", "heap", "unitCare"]),
            # The following requests are unused as they have an unknown purpose:
            # ("getBlock", None),
            # ("getBreakPoint", None),
        ]

        for (command, params) in requests:
            resp = self.exc_command(command, params)
            self.handle_command(command, resp["body"], resp["header"])

    def handle_mqtt_message(self, command, body, log=True):
        if log:
            LOGGER.debug("mq: {} {}".format(command, body))

        data = body["data"]
        if command == "onBattery":
            self.state.battery_level = data["value"]
        elif command == "onChargeState":
            self.state.is_charging = bool(data["isCharging"])
        elif command == "onCleanCount":
            self.state.clean_count = data["count"]
        elif command == "onCleanInfo_V2":
            if data["state"] == "clean":
                clean_state = data["cleanState"]
                motion_state = clean_state["motionState"]
                clean_type = clean_state["content"]["type"]

                if motion_state == "working":
                    self.state.state = VacuumState.RobotState.CLEANING
                elif motion_state == "pause":
                    self.state.state = VacuumState.RobotState.PAUSED
                else:
                    LOGGER.warning("Unhandled motion state: {}".format(motion_state))

                if clean_type == "auto":
                    self.state.clean_type = VacuumState.CleanType.AUTO
                elif clean_type == "spotArea":
                    self.state.clean_type = VacuumState.CleanType.SPOT_AREA
                elif clean_type == "customArea":
                    self.state.clean_type = VacuumState.CleanType.CUSTOM_AREA
                else:
                    LOGGER.warning("Unhandled clean type: {}".format(clean_type))

            elif data["state"] == "idle":
                self.state.state = VacuumState.RobotState.IDLE
                self.state.clean_type = None
            elif data["state"] == "goCharging":
                self.state.state = VacuumState.RobotState.RETURNING
                self.state.clean_type = None
        elif command == "onCleanPreference":
            self.state.cleaning_preference_enabled = bool(data["enable"])
        elif command == "onFwBuryPoint":
            cmd_payload = json.loads(data["content"])
            cmd_rn = cmd_payload["rn"]

            if cmd_rn == "bd_setting":
                cmd_data_raw = cmd_payload["d"]["body"]["data"]["d_val"]
                cmd_data = json.loads(cmd_data_raw)
                # Set normalized water level, water levels reported via
                # onFwBuryPoint::bd_setting are indexed from 0 through 3
                water_flow_map = {
                    0: VacuumState.WaterFlow.LOW,
                    1: VacuumState.WaterFlow.MEDIUM,
                    2: VacuumState.WaterFlow.HIGH,
                    3: VacuumState.WaterFlow.ULTRA_HIGH,
                }
                self.state.water_level = water_flow_map[cmd_data["waterAmount"]]
                self.state.auto_boost_suction_enabled = bool(cmd_data["isPressurized"])
            else:
                LOGGER.debug(
                    "Unhandled onFwBuryPoint message: {} {}".format(cmd_rn, cmd_payload)
                )
        elif command == "onSpeed":
            speed_map = {
                1000: VacuumState.Speed.QUIET,
                0: VacuumState.Speed.STANDARD,
                1: VacuumState.Speed.MAX,
                2: VacuumState.Speed.MAX_PLUS,
            }
            self.state.vacuum_speed = speed_map[data["speed"]]
        elif command == "onStats":
            # mq: onStats {'data': {'area': 0, 'time': 0, 'cid': '111', 'start': '1624877148', 'type': 'customArea', 'enablePowerMop': 0, 'powerMopType': 1, 'aiopen': 1, 'aitypes': [], 'avoidCount': 0}} # noqa: E501
            # mq: onStats {'data': {'area': 0, 'time': 0, 'cid': '111', 'start': '1624877304', 'type': 'spotArea', 'enablePowerMop': 0, 'powerMopType': 1, 'aiopen': 1, 'aitypes': [], 'avoidCount': 0}} # noqa: E501
            # mq: onStats {'data': {'area': 0, 'time': 0, 'cid': '111', 'start': '1624877619', 'type': 'auto', 'enablePowerMop': 0, 'powerMopType': 1, 'aiopen': 1, 'aitypes': [], 'avoidCount': 0}} # noqa: E501
            # mq: onStats {'data': {'area': 0, 'time': 0, 'cid': '111', 'start': '1624877619', 'type': 'auto', 'enablePowerMop': 0, 'powerMopType': 1, 'aiopen': 1, 'aitypes': [], 'avoidCount': 0}} # noqa: E501
            # mq: onStats {'data': {'area': 0, 'time': 61, 'cid': '111', 'start': '1624877304', 'type': 'spotArea', 'enablePowerMop': 0, 'powerMopType': 1, 'aiopen': 1, 'aitypes': [], 'avoidCount': 0}} # noqa: E501
            # TODO(NW): Handle mopping info in this message (enablePowerMop, powerMopType)
            self.state.clean_stats = CleanStats(
                area=data["area"],
                time=data["time"],
                avoid_count=data["avoidCount"],
                start_time=data["start"],
            )

            # TODO(NW): Determine whether clean type should be inferred from this?
            #             - does it introduce a race?
            typ = data["type"]
            if typ == "auto":
                self.state.clean_type = VacuumState.CleanType.AUTO
            elif typ == "spotArea":
                self.state.clean_type = VacuumState.CleanType.SPOT_AREA
            elif typ == "customArea":
                self.state.clean_type = VacuumState.CleanType.CUSTOM_AREA
            else:
                LOGGER.warning("Unknown clean type: {}".format(typ))
        elif command == "onTrueDetect":
            self.state.true_detect_enabled = bool(data["enable"])
        elif command == "onWaterInfo":
            self.state.mop_attached = bool(data["enable"])
            # Set normalized water level, water levels reported via onWaterInfo
            # are indexed from 1 through 4
            water_flow_map = {
                1: VacuumState.WaterFlow.LOW,
                2: VacuumState.WaterFlow.MEDIUM,
                3: VacuumState.WaterFlow.HIGH,
                4: VacuumState.WaterFlow.ULTRA_HIGH,
            }
            self.state.water_level = water_flow_map[data["amount"]]
        elif command == "reportStats":
            # mq: reportStats {'data': {'cid': '1117230632', 'stop': 0, 'enablePowerMop': 0, 'powerMopType': 2, 'stopReason': 1, 'startReason': 1, 'type': 'spotArea'}} # noqa: E501
            # mq: reportStats {'data': {'cid': '1117230632', 'stop': 1, 'enablePowerMop': 0, 'powerMopType': 1, 'stopReason': 2, 'startReason': 1, 'type': 'spotArea', 'mapCount': 9, 'area': 0, 'start': '1624877304', 'time': 61, 'content': '1', 'aiopen': 1, 'aitypes': [], 'aiavoid': 0}} # noqa: E501
            # mq: reportStats {'data': {'cid': '2132509283', 'stop': 0, 'enablePowerMop': 0, 'powerMopType': 2, 'stopReason': 1, 'startReason': 1, 'type': 'customArea'}} # noqa: E501
            # mq: reportStats {'data': {'cid': '2132509283', 'stop': 1, 'enablePowerMop': 0, 'powerMopType': 1, 'stopReason': 1, 'startReason': 1, 'type': 'customArea', 'mapCount': 9, 'area': 0, 'start': '1624877148', 'time': 0, 'content': '-2382.000000,-563.000000,-1998.000000,-1323.000000', 'aiopen': 1, 'aitypes': [], 'aiavoid': 0}} # noqa: E501
            # mq: reportStats {'data': {'cid': '2147037274', 'stop': 0, 'enablePowerMop': 0, 'powerMopType': 2, 'stopReason': 1, 'startReason': 1, 'type': 'spotArea'}} # noqa: E501
            # mq: reportStats {'data': {'cid': '67289670', 'stop': 0, 'enablePowerMop': 0, 'powerMopType': 2, 'stopReason': 2, 'startReason': 1, 'type': 'auto'}} # noqa: E501
            pass
        elif command == "onBreakPointStatus":
            pass
        elif command == "onCachedMapInfo":
            pass
        elif command in (
            "reportMinorMap",
            "reportPos",
            "reportMapTrace",
            "reportMapSubSet",
        ):
            pass
        elif command == "onError":
            pass
        elif command == "onEvt":
            pass
        elif command == "onMajorMap":
            pass
        elif command == "onMapSet":
            pass
        elif command == "onMapState":
            pass
        elif command == "onMapTrace":
            pass
        elif command == "onMinorMap":
            pass
        elif command == "onPos":
            pass
        elif command == "onRelocationState":
            pass
        elif command == "onRosNodeReady":
            pass
        elif command == "onSched_V2":
            pass
        else:
            LOGGER.warning("Unhandled mqtt command: {} {}".format(command, data))

    def exc_command(self, command, data=None):
        try:
            rv = self._api_client.exc_command(self._device, command, data)
        except ApiErrorException as e:
            self._err_count += 1
            if self._err_count >= 2:
                self.state.is_online = False
            raise e

        self._err_count = 0
        self.state.is_online = True
        return rv

    def handle_command(self, command: str, resp, header):
        if header is not None and "fwVer" in header:
            self.state.firmware_version = header["fwVer"]
        if header is not None and "hwVer" in header:
            self.state.hardware_version = header["hwVer"]

        if "data" not in resp:
            LOGGER.warning("No data provided for command: %s", command)
            return

        data = resp["data"]
        LOGGER.debug("http: {} {}".format(command, data))

        if command in (
            "getBattery",
            "getChargeState",
            "getCleanCount",
            "getCleanInfo_V2",
            "getCleanPreference",
            "getError",
            "getSpeed",
            "getStats",
            "getTrueDetect",
            "getWaterInfo",
        ):
            # Handle overlapping commands via MQTT routines
            mq_command = "on" + command.lstrip("get")
            self.handle_mqtt_message(mq_command, resp, log=False)

        elif command == "getInfo":
            for key, value in data.items():
                self.handle_command(key, value, None)
        elif command == "getTotalStats":
            self.state.total_stats = TotalStats(
                area=data["area"],
                time=data["time"],
                count=data["count"],
            )
        elif command == "getLifeSpan":
            self.state.lifespan = [
                ComponentLifeSpan(component=x["type"], left=x["left"], total=x["total"])
                for x in data
            ]
        elif command == "getCarpertPressure":
            self.state.auto_boost_suction_enabled = bool(data["enable"])
        elif command == "getAutoEmpty":
            self.state.auto_empty_enabled = bool(data["enable"])
        else:
            LOGGER.warning("Unhandled http command: {} {}".format(command, data))

    def _handle_state_change(self, state: VacuumState, attribute: str) -> None:
        for subscriber in self._subscribers:
            subscriber(state, attribute)

    def _start_polling(self):
        def poll_task():
            LOGGER.debug("Starting polling for %s", self._device.id)
            while self._should_poll:
                try:
                    self.force_refresh()
                except Exception:
                    LOGGER.exception("Error whilst polling, robot might be offline?")
                time.sleep(60 * 2)

            LOGGER.debug("Poll task for %s exiting", self._device.id)

        self._should_poll = True
        threading.Thread(target=poll_task, daemon=True).start()

    def _stop_polling(self):
        LOGGER.debug("Stopping polling for %s", self._device.id)
        self._should_poll = False

    def subscribe(self, handler):
        with self._lock:
            len_before = len(self._subscribers)
            self._subscribers.add(handler)
            if len_before == 0:
                # No subscribers! Subscribe and begin polling!
                self._subs_client.subscribe(self._device, self.handle_mqtt_message)
                self._start_polling()

    def unsubscribe(self, handler):
        with self._lock:
            self._subscribers.remove(handler)
            if len(self._subscribers) == 0:
                # No subscribers left! Stop polling and unsubscribe
                self._stop_polling()
                self._subs_client.unsubscribe(self._device, self.handle_mqtt_message)

    def set_true_detect(self, enabled: bool):
        self.exc_command(
            "setTrueDetect",
            {
                "enable": int(enabled),
            },
        )

    def set_clean_preference(self, enabled: bool):
        self.exc_command(
            "setCleanPreference",
            {
                "enable": int(enabled),
            },
        )

    def set_clean_count(self, count: bool):
        self.exc_command(
            "setCleanCount",
            {
                "count": int(count),
            },
        )

    def set_auto_empty(self, enabled: bool):
        self.exc_command(
            "setAutoEmpty",
            {
                "enable": int(enabled),
            },
        )
        # Optimistically update state, since no MQTT confirmation is received
        # for this setter.
        self.state.auto_empty_enabled = enabled

    def set_auto_boost_suction(self, enabled: bool):
        self.exc_command(
            "setCarpertPressure",
            {
                "enable": int(enabled),
            },
        )

    def set_water_level(self, level: VacuumState.WaterFlow):
        # TODO(NW): Colocate with deserialization and definition.
        water_level_map = {
            VacuumState.WaterFlow.LOW: 1,
            VacuumState.WaterFlow.MEDIUM: 2,
            VacuumState.WaterFlow.HIGH: 3,
            VacuumState.WaterFlow.ULTRA_HIGH: 4,
        }
        self.exc_command(
            "setWaterInfo",
            {
                "amount": water_level_map[level],
            },
        )

    def set_vacuum_speed(self, speed: VacuumState.Speed):
        # TODO(NW): Colocate with deserialization and definition.
        speed_map = {
            VacuumState.Speed.QUIET: 1000,
            VacuumState.Speed.STANDARD: 0,
            VacuumState.Speed.MAX: 1,
            VacuumState.Speed.MAX_PLUS: 2,
        }
        self.exc_command(
            "setSpeed",
            {
                "speed": speed_map[speed],
            },
        )

    def clean(self):
        self.exc_command(
            "clean_V2",
            {
                "act": "start",
                "content": {"count": "", "donotClean": "", "type": "auto", "value": ""},
                "mode": "",
                "router": "plan",
            },
        )

    def clean_areas(self, areas: List[int]):
        # TODO(NW): Use dynamic "count" for clean count.
        self.exc_command(
            "clean_V2",
            {
                "act": "start",
                "content": {
                    "count": "",
                    "donotClean": "",
                    "type": "spotArea",
                    "value": ",".join(str(x) for x in areas),
                },
                "mode": "",
                "router": "plan",
            },
        )

    def clean_custom(self, custom_area: str):
        # TODO(NW): Use dynamic "count" for clean count.
        self.exc_command(
            "clean_V2",
            {
                "act": "start",
                "content": {
                    "count": "",
                    "donotClean": "",
                    "type": "customArea",
                    "value": custom_area,
                },
                "mode": "",
                "router": "plan",
            },
        )

    def stop(self):
        # TODO(NW): See if the payload can be simplified (ala pause/resume)
        self.exc_command(
            "clean_V2",
            {
                "act": "stop",
                "content": {"count": "", "donotClean": "", "type": "", "value": ""},
                "mode": "",
                "router": "plan",
            },
        )

    def pause(self):
        self.exc_command("clean_V2", {"act": "pause"})

    def resume(self):
        self.exc_command("clean_V2", {"act": "resume"})

    def return_to_charge(self):
        self.exc_command("charge", {"act": "go"})

    def relocate(self):
        self.exc_command("setRelocationState", {"mode": "manu"})

    def play_sound(self, sound_id=30):
        self.exc_command("playSound", {"count": 1, "sid": sound_id})



from __future__ import annotations

import logging
import signal
import time
from dataclasses import dataclass, asdict

import click
from terminaltables import AsciiTable

from deebot_t8 import (
    ApiClient,
    DeebotEntity,
    PortalClient,
    DeebotAuthClient,
    SubscriptionClient,
    Credentials,
)
from deebot_t8.auth_client import Authenticator
from deebot_t8.entity import VacuumState
from deebot_t8.md5 import md5_hex
from .config import Config, load_config, write_config

LOGGER = logging.getLogger(__name__)


@dataclass
class TypedObj:
    config_path: str
    config: Config

    auth_client: DeebotAuthClient
    api_client: ApiClient | None = None
    subscription_client: SubscriptionClient | None = None

    entity: DeebotEntity | None = None


@click.group()
@click.pass_context
@click.option(
    "--config-file", type=click.Path(exists=False), default="deebot.conf.json"
)
def cli(ctx, config_file):
    logging.basicConfig()
    logging.getLogger("deebot_t8").setLevel(logging.DEBUG)

    config = load_config(config_file)

    def on_credentials_changed(credentials: Credentials):
        config.credentials = credentials
        write_config(config_file, config)

    auth_client = None
    api_client = None
    subscription_client = None

    if config is not None:
        portal_client = PortalClient(
            device_id=config.device_id,
            country=config.country,
            continent=config.continent,
        )
        auth_client = DeebotAuthClient(
            portal_client=portal_client,
            device_id=config.device_id,
            country=config.country,
        )
        authenticator = Authenticator(
            auth_client=auth_client,
            country=config.country,
            device_id=config.device_id,
            account_id=config.username,
            password_hash=config.password_hash,
            cached_credentials=config.credentials,
            on_credentials_changed=on_credentials_changed,
        )
        api_client = ApiClient(portal_client=portal_client, authenticator=authenticator)
        subscription_client = SubscriptionClient(
            authenticator=authenticator,
            continent=config.continent,
            device_id=config.device_id,
        )

    ctx.obj = TypedObj(
        config_path=config_file,
        config=config,
        auth_client=auth_client,
        api_client=api_client,
        subscription_client=subscription_client,
    )


def renew_access_tokens_impl(auth: DeebotAuthClient, config: Config):
    return auth.login(config.username, config.password_hash)


@cli.command()
@click.pass_obj
@click.option("--username", type=str, required=True)
@click.option("--password", type=str, required=True)
@click.option("--country", type=str, required=True)
@click.option("--continent", type=str, required=True)
@click.option("--regen-device", type=bool)
def login(obj: TypedObj, username, password, country, continent, regen_device):
    # TODO(NW): Infer continent from country

    if not regen_device and obj.config is not None and obj.config.device_id is not None:
        # Reuse existing device id if one exists
        device_id = obj.config.device_id
    else:
        device_id = md5_hex(str(time.time()))

    obj.config = Config(
        username=username,
        password_hash=md5_hex(password),
        device_id=device_id,
        country=country,
        continent=continent,
    )
    # Recreate clients for this special use case to apply new configuration
    # parameters (country, continent, device id)
    portal_client = PortalClient(
        device_id=obj.config.device_id,
        country=obj.config.country,
        continent=obj.config.continent,
    )
    auth_client = DeebotAuthClient(
        portal_client=portal_client,
        device_id=obj.config.device_id,
        country=obj.config.country,
    )

    write_config(obj.config_path, obj.config)
    obj.config.credentials = renew_access_tokens_impl(auth_client, obj.config)
    write_config(obj.config_path, obj.config)

    if obj.config.credentials is None:
        raise AssertionError

    click.echo(
        "Authenticated with user {}, token expires at {}".format(
            obj.config.credentials.user_id,
            obj.config.credentials.expires_at,
        )
    )


@cli.command()
@click.pass_obj
def renew_access_token(obj: TypedObj):
    obj.config.credentials = renew_access_tokens_impl(obj.auth_client, obj.config)
    write_config(obj.config_path, obj.config)

    if obj.config.credentials is None:
        raise AssertionError

    click.echo(
        "Renewed with user {}, token expires at {}".format(
            obj.config.credentials.user_id,
            obj.config.credentials.expires_at,
        )
    )


@cli.command()
@click.pass_obj
def list_devices(obj: TypedObj):
    if obj.api_client is None:
        raise AssertionError

    devices = obj.api_client.get_devices_list()
    table_data = [
        ["device id", "name", "product category", "model", "status"],
    ]
    for device in devices:
        table_data.append(
            [
                device.id,
                device.name,
                device.product_category,
                device.model_name,
                # status 0 seems to indicate offline?
                # status 1 online
                device.status,
            ]
        )
    print(AsciiTable(table_data).table)


@cli.group()
@click.pass_obj
@click.argument("device-name", type=str, required=False)
def device(obj: TypedObj, device_name):
    if obj.api_client is None or obj.subscription_client is None:
        raise AssertionError

    selected_device = None
    devices = obj.api_client.get_devices_list()
    for d in devices:
        if d.name == device_name:
            selected_device = d
            break
    else:
        raise click.UsageError(
            "Device with specified name ({}) could not be found.".format(device_name)
        )

    obj.entity = DeebotEntity(
        api_client=obj.api_client,
        subs_client=obj.subscription_client,
        device=selected_device,
    )


@device.command()
@click.pass_obj
def subscribe(obj: TypedObj):
    # Silence the logger to allow our table to display nicely
    logging.getLogger("deebot_t8").setLevel(logging.ERROR)
    if obj.entity is None:
        raise AssertionError

    def on_state_change(state: VacuumState, attribute: str):
        click.clear()
        table_data = [["Attribute", "Value"]]
        for attr, value in asdict(state).items():
            if isinstance(value, list):
                for entry in value:
                    table_data.append([attr, entry])
            else:
                table_data.append([attr, value])
        print(AsciiTable(table_data).table)

    obj.entity.subscribe(on_state_change)

    def handler(signum, frame):
        obj.entity.unsubscribe(on_state_change)
        raise KeyboardInterrupt

    # Wait forever (until SIGINT at least)
    signal.signal(signal.SIGINT, handler)
    signal.pause()


@device.command()
@click.pass_obj
def clean(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.clean()


@device.command()
@click.pass_obj
@click.argument("areas", type=int, nargs=-1, required=True)
def clean_areas(obj: TypedObj, areas):
    if obj.entity is None:
        raise AssertionError

    obj.entity.clean_areas(areas)


@device.command()
@click.pass_obj
@click.argument("custom", type=str)
def clean_custom(obj: TypedObj, custom):
    if obj.entity is None:
        raise AssertionError

    obj.entity.clean_custom(custom)


@device.command()
@click.pass_obj
def stop(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.stop()


@device.command()
@click.pass_obj
def pause(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.pause()


@device.command()
@click.pass_obj
def return_to_charge(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.return_to_charge()


@device.command()
@click.pass_obj
def relocate(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.relocate()


@device.command()
@click.pass_obj
def play_sound(obj: TypedObj):
    if obj.entity is None:
        raise AssertionError

    obj.entity.play_sound()


@device.command()
@click.pass_obj
@click.argument("enable", type=bool)
def set_true_detect(obj: TypedObj, enable: bool):
    if obj.entity is None:
        raise AssertionError

    obj.entity.set_true_detect(enable)


@device.command()
@click.pass_obj
@click.argument("enable", type=bool)
def set_clean_preference(obj: TypedObj, enable: bool):
    if obj.entity is None:
        raise AssertionError

    obj.entity.set_clean_preference(enable)


@device.command()
@click.pass_obj
def set_water_level(obj: TypedObj):
    pass


@device.command()
@click.pass_obj
def set_vacuum_speed(obj: TypedObj):
    pass


@device.command()
@click.pass_obj
def send_command(obj: TypedObj):
    pass


if __name__ == "__main__":
    cli()



import os

from setuptools import setup

readme_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "README.md",
)
long_description = open(readme_path).read()

setup(
    name="deebot-t8",
    packages=["deebot_t8", "deebot_t8.cli"],
    author="Nick Whyte",
    author_email="nick@nickwhyte.com",
    description="Yet another Ecovacs Deebot API client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nickw444/deebot-t8",
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
    ],
    install_requires=[
        "requests>=2.25.1,<3.0.0",
        'dataclasses;python_version<"3.7"',
        "paho-mqtt>=1.5.1,<2.0.0",
    ],
    extras_require={"cli": ["click>=8.0.1,<9.0.0", "terminaltables>=3.1.0,<4.0.0"]},
    entry_points={
        "console_scripts": ["deebot-t8=deebot_t8.cli.__main__:cli"],
    },
)



from __future__ import annotations

import logging
import threading
import time
from typing import Dict, Any, Callable

import requests

from .credentials import Credentials
from .exceptions import InvalidCredentialsException, ApiErrorException
from .md5 import md5_hex
from .portal_client import PortalClient
from .urls import USER_DO_PATH, REALM

LOGGER = logging.getLogger(__name__)

CLIENT_KEY = "1520391301804"
CLIENT_SECRET = "6c319b2a5cd3e66e39159c2e28f2fce9"
AUTH_CLIENT_KEY = "1520391491841"
AUTH_CLIENT_SECRET = "77ef58ce3afbe337da74aa8c5ab963a9"


class DeebotAuthClient:
    def __init__(self, portal_client: PortalClient, device_id: str, country: str):
        self._portal_client = portal_client
        self._device_id = device_id
        self._country = country
        self._meta = {
            "country": country,
            "lang": "EN",
            "deviceId": device_id,
            "appCode": "global_e",
            "appVersion": "1.6.3",
            "channel": "google_play",
            "deviceType": "1",
        }

    def _sign_params(
        self,
        params: Dict[Any, Any],
        client_key: str,
        client_secret: str,
    ):
        payload = (
            client_key
            + "".join([k + "=" + str(params[k]) for k in sorted(params.keys())])
            + client_secret
        )
        return md5_hex(payload)

    def _get_login_url(self):
        tld = "cn" if self._country == "cn" else "com"
        login_path = "user/loginCheckMobile" if self._country == "cn" else "user/login"
        return (
            "https://gl-{country}-api.ecovacs.{tld}/v1/private/{country}/"
            "{lang}/{deviceId}/{appCode}/{appVersion}/{channel}/"
            "{deviceType}/{login_path}"
        ).format(login_path=login_path, tld=tld, **self._meta)

    def _get_authcode_url(self):
        tld = "cn" if self._country == "cn" else "com"
        return (
            f"https://gl-{self._country}-openapi.ecovacs.{tld}/"
            f"v1/global/auth/getAuthCode"
        )

    def do_account_password_exchange(
        self, account_id: str, password_hash: str
    ) -> Credentials:
        params = {
            "requestId": md5_hex(str(time.time())),
            "account": account_id,
            "password": password_hash,
            "authTimespan": int(time.time() * 1000),
            "authTimeZone": "GMT-8",
        }

        # Sign params
        params_sig = self._sign_params(
            {**self._meta, **params}, CLIENT_KEY, CLIENT_SECRET
        )
        params["authSign"] = params_sig
        params["authAppkey"] = CLIENT_KEY

        url = self._get_login_url()

        # Do request
        resp = requests.get(url, params)
        resp.raise_for_status()
        resp_json = resp.json()
        if resp_json["code"] == "0000":
            return Credentials(
                access_token=resp_json["data"]["accessToken"],
                user_id=resp_json["data"]["uid"],
                expires_at=None,
            )
        elif resp_json["code"] in ("1005", "1010"):
            raise InvalidCredentialsException("Invalid email or password")
        else:
            raise Exception(
                "Unknown error: {} ({})".format(
                    resp_json["msg"],
                    resp_json["code"],
                )
            )

    def do_get_authcode(self, uid: str, access_token: str):
        params: Dict[str, str | int] = {
            "uid": uid,
            "accessToken": access_token,
            "bizType": "ECOVACS_IOT",
            "deviceId": self._device_id,
            "authTimespan": int(time.time() * 1000),
        }

        # Sign params
        params_sig = self._sign_params(
            {
                "openId": "global",
                **params,
            },
            AUTH_CLIENT_KEY,
            AUTH_CLIENT_SECRET,
        )
        params["authSign"] = params_sig
        params["authAppkey"] = AUTH_CLIENT_KEY

        # Do request
        resp = requests.get(self._get_authcode_url(), params=params)
        resp.raise_for_status()
        resp_json = resp.json()

        if resp_json["code"] == "0000":
            return resp_json["data"]["authCode"]
        elif resp_json["code"] == "1005":
            raise InvalidCredentialsException("Invalid email or password")
        else:
            raise ApiErrorException(
                "Unknown error: {} ({})".format(
                    resp_json["msg"],
                    resp_json["code"],
                )
            )

    def do_login_by_iot_token(self, user_id: str, auth_code: str):
        org = "ECOCN" if self._country == "cn" else "ECOWW"
        country = "Chinese" if self._country == "cn" else self._country.upper()

        resp = self._portal_client.do_post(
            USER_DO_PATH,
            {
                "todo": "loginByItToken",
                "edition": "ECOGLOBLE",
                "userId": user_id,
                "token": auth_code,
                "realm": REALM,
                "resource": self._device_id[0:8],
                "org": org,
                "last": "",
                "country": country,
            },
        )

        if resp["result"] == "ok":
            return Credentials(
                access_token=resp["token"],
                user_id=resp["userId"],
                # Tokens appear to have ~7 day expiry.
                # Set expiry to 2 days to eagerly refresh
                # TODO(NW): Decode the JWT header returned and pass along the
                #  expiry in this field
                expires_at=int(time.time()) + 60 * 60 * 24 * 2,
            )

        raise ApiErrorException("Unknown error: {}".format(resp))

    def login(self, account_id: str, password_hash: str):
        exch_resp = self.do_account_password_exchange(account_id, password_hash)
        auth_code = self.do_get_authcode(exch_resp.user_id, exch_resp.access_token)
        return self.do_login_by_iot_token(exch_resp.user_id, auth_code)


class Authenticator:
    def __init__(
        self,
        auth_client: DeebotAuthClient,
        country: str,
        device_id: str,
        account_id: str,
        password_hash: str,
        cached_credentials: Credentials = None,
        on_credentials_changed: Callable[[Credentials], None] = None,
    ):
        self._auth_client = auth_client
        self._country = country
        self._device_id = device_id

        self._account_id = account_id
        self._password_hash = password_hash

        self._lock = threading.Lock()
        self._credentials = cached_credentials
        self._on_credentials_changed = on_credentials_changed

    def authenticate(self, force=False):
        with self._lock:
            should_login = False
            if self._credentials is None or force:
                LOGGER.debug("No cached credentials, performing login")
                should_login = True
            elif self._credentials.expires_at < time.time():
                LOGGER.debug("Credentials have expired, performing login")
                should_login = True

            if should_login:
                self._credentials = self._auth_client.login(
                    self._account_id, self._password_hash
                )

                if self._on_credentials_changed is not None:
                    self._on_credentials_changed(self._credentials)

            return self._credentials

    def invalidate(self):
        pass
